/*-----------------------------------------------------------------------

File  : che_proofcontrol.c

Author: Stephan Schulz

Contents
 
  Basic functions for proof control objects.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Fri Oct 16 16:20:00 MET DST 1998
    Salvaged from cco_proofstate.h, forked control and state.
<2> Wed Dec 16 18:45:14 MET 1998
    Moved from cco to che

-----------------------------------------------------------------------*/

#include "che_proofcontrol.h"


/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

#define CHE_PROOFCONTROL_INTERNAL

char* DefaultWeightFunctions = 
"\n"
"weight21_ugg  = Clauseweight(PreferUnitGroundGoals,2,1,1)      \n"
"rweight21_g   = Refinedweight(PreferGoals,    2,1,1.5,1.1,1) \n"
"rweight11_g   = Refinedweight(PreferGoals,    1,1,1.5,1.1,1.1) \n"
"rweight21_a   = Refinedweight(PreferNonGoals, 2,1,1.5,1.1,1.1) \n"
"rweight21_ugg = Refinedweight(PreferUnitGroundGoals, 2,1,1.5,1.1,1.1) \n"
"fifo_f        = FIFOWeight(ConstPrio)                      \n"
"lifo_f        = LIFOWeight(ConstPrio)                      \n"
"weight11_f    = Clauseweight(ConstPrio,1,1,1)              \n"
"weight21_f    = Clauseweight(ConstPrio,2,1,1)              \n"
;

char* DefaultHeuristics = 
"Weight     = (1*weight21_ugg)                       \n"
"RWeight    = (1*rweight21_ugg)                      \n"
"FIFO       = (1*fifo_f)                             \n"
"LIFO       = (1*lifo_f)                             \n"
"Default    = (3*rweight21_a, 1*rweight21_g)         \n"
"Uniq       = (1*Uniqweight(ConstPrio))\n"
"/* Here start the auto-mode-heuristics*/            \n"

"/* H*-Strategies: */  \n"

#include "che_H_____auto.c"

"/* G* Strategies: */  \n"

#include "che_G_____auto.c"

"/* U* Strategies */ \n"

#include "che_U_____auto.c"

;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ProofControlFree()
//
//   Allocate an empty, initialized ProofControlCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

ProofControl_p ProofControlAlloc()
{
   ProofControl_p handle = ProofControlCellAlloc();
   handle->ocb                     = NULL;
   handle->wfcbs                   = WFCBAdminAlloc();
   handle->hcbs                    = HCBAdminAlloc();
   handle->hcb                     = NULL;
   handle->prefer_initial_clauses  = false;
   handle->select_on_proc_only     = false;
   handle->selection_strategy      = SelectNoLiterals;
   handle->pos_lit_sel_min         = 0; 
   handle->pos_lit_sel_max         = LONG_MAX; 
   handle->neg_lit_sel_min         = 0; 
   handle->neg_lit_sel_max         = LONG_MAX; 
   handle->all_lit_sel_min         = 0; 
   handle->all_lit_sel_max         = LONG_MAX; 
   handle->weight_sel_min          = 0;
   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: ProofControlFree()
//
//   Free a ProofControlCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ProofControlFree(ProofControl_p junk)
{
   assert(junk);
   if(junk->ocb)
   {
      OCBFree(junk->ocb);
   }
   WFCBAdminFree(junk->wfcbs);
   HCBAdminFree(junk->hcbs);
   /* hcb is always freed in junk->hcbs */
   ProofControlCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: DoLiteralSelection()
//
//   Based on control, select a literal selection strategy and apply
//   it to clause.
//
// Global Variables: -
//
// Side Effects    : Changes properties in clause
//
/----------------------------------------------------------------------*/

void DoLiteralSelection(ProofControl_p control, Clause_p clause)
{
   if((clause->pos_lit_no >= control->pos_lit_sel_min) && 
      (clause->pos_lit_no <= control->pos_lit_sel_max) &&
      (clause->neg_lit_no >= control->neg_lit_sel_min) && 
      (clause->neg_lit_no <= control->neg_lit_sel_max) &&
      (ClauseLiteralNumber(clause) >= control->all_lit_sel_min) &&
      (ClauseLiteralNumber(clause) <= control->all_lit_sel_max) &&
      ((control->weight_sel_min==0) || /* Efficiency hack - only
					 compute clause weight if this
					 option is activated */
       (control->weight_sel_min<=ClauseStandardWeight(clause))))
   {
      control->selection_strategy(clause);
   }
   else
   {
      SelectNoLiterals(clause); 
   }
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/







